// ----------------------------------------------------------------------------
// Instantiate.java
// The "instantiate" command.
// $Id: Instantiate.java,v 1.16 2006/04/03 16:22:39 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import org.w3c.dom.*;
import fmrisc.Proving.*;
import fmrisc.ProofNavigator.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;

/**
 * 
 * The "instantiate" command: instantiate a quantified formula.
 *
 */
public final class Instantiate extends ProofCommandBase
{
  public final static String Name = "instantiate";
  private String label;        // the label of the formula to be instantiated
  private Expression[] values; // the instantiation values
  
  /***************************************************************************
   * Create an "instantiate" command.
   * @param label the label of the formula to be instantiated
   * @param values the instantiation values
   **************************************************************************/
  public Instantiate(String label, Expression[] values)
  {
    super(Name);
    this.label = label;
    this.values = values;
  }
  
  /****************************************************************************
   * Process command "instantiate": 
   * instantiate universal assumption or existential goal.
   * @param state the proof state to which the command is applied
   * @return null or array of one or two new proof states.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Formula formula = state.getFormula(label);
    if (formula == null)
    {
      error("no formula with label " + label + " in current state");
      return null;
    }
    Expression exp = formula.getExpression();
    TypedIdentifier[] vars;
    Expression base;
    if (formula.isGoal())
    {  
      if (!(exp instanceof ExistsFormula))
      {
        error ("goal " + label + " cannot be instantiated " +
          "(existentially quantified formula expected)");
        return null;
      }
      ExistsFormula f = (ExistsFormula)exp;
      vars = f.getVariables();
      base = f.getBase();
    }
    else
    {
      if (!(exp instanceof ForallFormula))
      {
        error ("assumption " + label + " cannot be instantiated " +
          "(universally quantified formula expected)");
        return null;
      }
      ForallFormula f = (ForallFormula)exp;
      vars = f.getVariables();
      base = f.getBase();
    }
    
    int n = vars.length;
    if (n != values.length)
    {
      error ("wrong number of instantiation values (" +
        n + " expected, " + values.length + " given)");
      return null;
    }
    
    boolean isAnonym[] = new boolean[n];
    int anonyms = 0;
    
    Context context = State.getContext();
    context.open();
    Environment env = state.getEnvironment();
    for (int i=0; i<n; i++)
    {
      Expression value = values[i];
      isAnonym[i] = value instanceof Reference && 
        ((Reference)value).getName().equals("_");
      if (isAnonym[i]) 
      {
        anonyms = anonyms+1;
        continue;
      }
      Type type = typeCheck(env, value);
      if (type == null) 
      {
        context.abort();
        return null;
      }
      TypedIdentifier var = vars[i];
      Type type0 = var.getIdentifier().getSymbol().getType();
      if (!Checking.hasType(value, type, type0, true)) 
      {
        context.abort();
        error("type " + type.toString() + " of value " + value.toString() +
          " does not match variable declaration " + var.toString());
        return null;
      }
    }
    Expression tcc = context.getFormula();
    context.abort();
    
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    
    ProofState state1 = null;
    if (tcc != null)
    {
      // create child state to prove tcc
      state1 = new ProofState(state, env);
      state1.setAssumptions(state1.copy(assumptions));
      Formula goals0[] = { new Formula(state1, tcc, true) };
      state1.setGoals(goals0);
    }
    
    TypedIdentifier[] vars0 = new TypedIdentifier[anonyms];
    ValueDeclIdentifier[] vars1 = new ValueDeclIdentifier[n-anonyms];
    Expression[] values1 = new Expression[n-anonyms];
    int i0 = 0;
    int i1 = 0;
    for (int i=0; i<n; i++)
    {
      if (isAnonym[i])
      {
        vars0[i0] = vars[i].instantiate();
        vars[i].setSubstitution(vars0[i0]);
        i0 = i0+1;
      }
      else
      {
        vars1[i1] = vars[i].getIdentifier();
        values1[i1] = values[i];
        i1 = i1+1;
      }
    }
    Expression base0 = base.instantiate();
    for (int i=0; i<n; i++)
      vars[i].resetSubstitution(); 
    Expression exp0 = base0;
    if (anonyms > 0)
    {
      if (exp instanceof ForallFormula)
        exp0 = Construct.forallFormula(vars0, exp0);
      else
        exp0 = Construct.existsFormula(vars0, exp0);
    }
    if (anonyms < n)
      exp0 = Substitute.substitute(exp0, vars1, values1, false);
    
    ProofState state0 = new ProofState(state, state.getEnvironment());
    Formula formula0 = new Formula(state0, exp0, formula.isGoal());
    
    if (formula.isGoal())
    {
      // flip negation of existential formulas to assumptions
      Formula formula1 = new Formula(state0, 
          ProofUtil.negate(formula.getExpression()), false);
      state0.setAssumptions(state0.add(assumptions, formula1));
      state0.setGoals(state0.replace(goals, formula, formula0));
    }
    else
    {
      state0.setAssumptions(state0.add(assumptions, formula0));
      state0.setGoals(state0.copy(goals));
    }
    if (tcc != null)
      state0.setAxioms(state0.add(state0.getAxioms(),
          new Formula(state0, tcc, false)));
    if (state1 == null)
      return new ProofState[] { state0 };
    else
      return new ProofState[] { state0, state1 };
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    node.appendChild(newTextNode(label));
    int n = values.length;
    for (int i=0; i<n; i++)
      node.appendChild(toNode(values[i]));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n == 0)
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    String label = toString(children.item(0));
    if (label == null) return null;
    Expression[] values = new Expression[n-1];
    for (int i=1; i<n; i++)
    {
      values[i-1] = toExpression(children.item(i));
      if (values[i-1] == null) return null;
    }
    return new Instantiate(label, values);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    int n = values.length;
    for (int i=0; i<n; i++)
    {
      values[i].print(out);
      if (i+1 < n) out.print(", ");
    }
    out.print(" in ");
    out.print(label);
  }
}
