// ----------------------------------------------------------------------------
// ProofCommandBase.java
// Base class of proof commands
// $Id: ProofCommandBase.java,v 1.13 2007/10/23 20:09:59 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.SWT.*;
import fmrisc.Proving.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;

/**
 * 
 * Base class of proof commands.
 *
 */
public abstract class ProofCommandBase extends CommandBase 
  implements ProofCommand
{
  /***************************************************************************
   * construct proof command with denoted name.
   * @param name the name of the proof command
   **************************************************************************/
  public ProofCommandBase(String name)
  {
    super(name);
  }
  
  /****************************************************************************
   * process proof state (in environment denoted by that state)
   * @param state the state to which the proof command is applied 
   *        (remains unchanged)
   * @return a sequence of children of state generated by the command
   *         (null, if the command did not generate any)
   ***************************************************************************/
  protected abstract ProofState[] processCore(ProofState state);
  
  /****************************************************************************
   * process proof state (in environment denoted by that state).
   * This function rather than processCore() must be called by any 
   * proof command that wishes to execute another proof command.
   * @param state the state to which the proof command is applied 
   *        (remains unchanged)
   * @return a sequence of children of state generated by the command
   *         (null, if the command did not generate any)
   ***************************************************************************/
  public final ProofState[] process(ProofState state)
  {
    // we need to make sure that simplification is not switched off
    // in recursive invocation of command
    ProofState[] children = State.getReplayChildren();
    State.setReplayChildren(null);
    ProofState[] result = processCore(state);
    State.setReplayChildren(children);
    return result;
  }
  
  /****************************************************************************
   * apply command to current state of current proof.
   ***************************************************************************/
  public void process()
  {
    Proof proof = State.getProof();
    if (proof == null)
    {
      error("command only allowed in proof mode");
      return;
    }
    ProofState state = proof.getCurrent();
    if (state == null) 
    {
      error("proof is closed");
      return;
    }
    State.setAborted(false);
    Environment oenv = State.getEnvironment();
    State.setEnvironment(state.getEnvironment());
    boolean oldAuto = State.useAutoSimplify();
    State.setAutoSimplify(state.useAutoSimplify());
    ProofState[] children = processCore(state);
    State.setEnvironment(oenv);
    State.setAutoSimplify(oldAuto);
    // abort button may have been pressed
    // command redo may have been executed
    if (!State.isAborted() && state == proof.getCurrent()) 
    {
      if (children == null)
      {
        if (state.isClosed()) proof.close(state);
      }
      else 
      {  
        state.extend(this, children);
        if (State.isAborted()) proof.cut(state);
      }
    }
    MainSWT.updateMenu();
  }
  
  /**************************************************************************
   * Typecheck formula in environment; 
   * print error message if something goes wrong.
   * Be careful: may generate a type checking condition as a side-effect!
   * @param env an environment
   * @param exp expression to be type checked in env
   * @return true iff exp is a valid formula (i.e. has result type BOOLEAN).
   *************************************************************************/
  protected boolean typeCheckFormula(Environment env, Expression exp)
  {
    Type type = typeCheck(env, exp);
    if (type == null) return false;
    if (!Checking.isBoolean(type))
    {
      error("expression has type " + type.toString() + " (BOOLEAN expected)");
      return false;
    }
    return true;
  }
  
  /**************************************************************************
   * Typecheck expression in environment; 
   * print error message if something goes wrong.
   * Be careful: may generate a type checking condition as a side-effect!
   * @param env an environment
   * @param exp expression to be type checked in env
   * @return type of exp (null, if something went wrong).
   *************************************************************************/
  protected Type typeCheck(Environment env, Expression exp)
  {
    Environment oenv = State.getEnvironment();
    State.setEnvironment(env);
    Type type = Checking.process(exp);
    State.setEnvironment(oenv);
    return type;
  }
}
