// ----------------------------------------------------------------------------
// Checking.java
// Semantics (type) checking routines.
// $Id: Checking.java,v 1.23 2009/10/29 12:07:23 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Semantics;

import java.util.*;
import fmrisc.ProofNavigator.*;
import fmrisc.Syntax.*;

/**
 * 
 * type checking routines
 *
 */
public final class Checking
{
  // canonical type tables
  private static Vector btable; // basic (atomic) types
  private static Vector ftable; // function types
  private static Vector ttable; // tuple types
  private static Vector rtable; // record types
  private static Vector atable; // array types
  private static Vector stable; // subtypes
  private static Vector utable; // subrange types
  
  // canonical typed identifier table
  private static Vector titable; // typed identifiers

  // types used in type checking
  private static Type booleanType;
  private static Type intType;
  private static Type natType;
  private static Type realType;
  private static Type real0Type;
  
  // true iff subtype matching is disabled for function types
  // strengthens type discipline
  private static boolean strongerTypes;
  
  // true iff INT matches NAT
  // weakens type discipline
  private static boolean weakerTypes;
  
  // collection of symbols touched during type-checking
  private static Collection referencedSymbols;
  
  /***************************************************************************
   * Initialize checking state
   **************************************************************************/
  public static void init()
  {
    // canonical type tables
    btable = new Vector(); // basic (atomic) types
    ftable = new Vector(); // function types
    ttable = new Vector(); // tuple types
    rtable = new Vector(); // record types
    atable = new Vector(); // array types
    stable = new Vector(); // subtypes
    utable = new Vector(); // subrange types
    
    // canonical typed identifier table
    titable = new Vector(); // typed identifiers
    
    // types used in type checking
    booleanType = canonical(Construct.atomicType("BOOLEAN"));
    intType  = canonical(Construct.atomicType("INT"));
    natType = canonical(Construct.atomicType("NAT"));
    realType = canonical(Construct.atomicType("REAL"));
    real0Type  = canonical(Construct.atomicType("REAL0"));
    
    // default type discipline
    strongerTypes = false;
    weakerTypes = false;
    
    // collection of symbols touched during type-checking
    referencedSymbols = new HashSet(); 
  }
  
  /****************************************************************************
   * signals whether stronger type discipline is to be preserved
   * @param stronger is true iff stronger type discipline is to be preserved
   **************************************************************************/
  public static void setStrongerTypes(boolean stronger)
  {
    strongerTypes = stronger;
  }
  
  /****************************************************************************
   * signals whether weaker type discipline is to be preserved
   * @param weaker is true iff weaker type discipline is to be preserved
   **************************************************************************/
  public static void setWeakerTypes(boolean weaker)
  {
    weakerTypes = weaker;
  }
  
  /****************************************************************************
   * Get canonical type BOOLEAN
   ***************************************************************************/
  public static Type getBooleanType()
  {
    return booleanType;
  }
  
  /****************************************************************************
   * Get canonical type INT
   ***************************************************************************/
  public static Type getIntType()
  {
    return intType;
  }
  /****************************************************************************
   * Get canonical type NAT
   ***************************************************************************/
  public static Type getNatType()
  {
    return natType;
  }
  
  /****************************************************************************
   * checks if type equals BOOLEAN
   * @param type
   * @return true iff type equals BOOLEAN
   ***************************************************************************/
  public static boolean isBoolean(Type type)
  {
    return type == booleanType;
  }
 
  /****************************************************************************
   * checks if type equals INT
   * @param type
   * @return true iff type equals INT
   ***************************************************************************/
  public static boolean isInt(Type type)
  {
    return type == intType;
  }
  
  /****************************************************************************
   * checks if type equals NAT
   * @param type
   * @return true iff type equals NAT
   ***************************************************************************/
  public static boolean isNat(Type type)
  {
    return type == natType;
  }
  
  /****************************************************************************
   * get collection of referenced symbols
   ***************************************************************************/
  public static Collection getReferencedSymbols()
  {
    return referencedSymbols;
  }
  
  /****************************************************************************
   * reset collection of referenced symbols
   ***************************************************************************/
  public static void resetReferencedSymbols()
  {
    referencedSymbols = new HashSet();
  }
  
  /****************************************************************************
   * returns canonical version of type (equality of two types implies
   * the equality of the canonical version of the types).
   * @param type the type
   * @return the canonical version of the type (null, if an error occured)
   **************************************************************************/
  private static Type canonical0(Type type)
  {
    if (type instanceof Reference)
    {
      Reference ident = (Reference)type;
      TypeSymbol symbol = State.getEnvironment().getTypeSymbol(ident);
      if (symbol == null)
      {
        State.setError("type " + ident.getName() + " is not declared");
        return null;
      }
      referencedSymbols.add(symbol);
      ident.setSymbol(symbol);
      return symbol.getType();
    }
    
    if (type instanceof AtomicType)
    {
      String name = ((AtomicType)type).getName();
      Enumeration e = btable.elements();
      while (e.hasMoreElements())
      {
        AtomicType ctype = (AtomicType)(e.nextElement());
        if (name.equals(ctype.getName())) return ctype;
      }
      btable.add(type);
      return type;
    }
    
    if (type instanceof FunctionType)
    {
      FunctionType type0 = (FunctionType)type;
      Type[] domain0 = type0.getDomain();
      int n = domain0.length;
      Type[] domain = new Type[n];
      for (int i=0; i<n; i++)
      {
        domain[i] = canonical0(domain0[i]);
        if (domain[i] == null) return null;
      }
      Type range = canonical0(type0.getRange());
      if (range == null) return null;
      Enumeration e = ftable.elements();
      L: while (e.hasMoreElements())
      {
        FunctionType ctype = (FunctionType)(e.nextElement());
        Type[] cdomain = ctype.getDomain();
        if (n != cdomain.length) continue L;
        for (int i=0; i<n; i++)
          if (domain[i] != cdomain[i]) continue L;
        if (range == ctype.getRange())
          return ctype;
      }
      type = new FunctionType(domain, range);
      ftable.add(type);
      return type;
    }
    
    if (type instanceof TupleType)
    {
      TupleType type0 = (TupleType)type;
      Type[] base = type0.getBase();
      int n = base.length;
      Type[] base0 = new Type[n];
      for (int i=0; i<n; i++)
        base0[i] = canonical0(base[i]);
      Enumeration e = ttable.elements();
      L: while (e.hasMoreElements())
      {
        TupleType ctype = (TupleType)(e.nextElement());
        Type[] cbase = ctype.getBase();
        if (n != cbase.length) continue;
        for (int i=0; i<n; i++)
          if (base0[i] != cbase[i]) continue L;
        return ctype;
      }
      type = new TupleType(base0);
      ttable.add(type);
      return type;
    }
    
    if (type instanceof RecordType)
    {
      RecordType type0 = (RecordType)type;
      TypedIdentifier[] base = type0.getBase();
      int n = base.length;
      for (int i=0; i<n; i++)
      {
        String name = base[i].getIdentifier().getName();
        for (int j=0; j<i; j++)
        {
          if (name.equals(base[j].getIdentifier().getName()))
          {
            State.setError("field name " + name + " occurs twice in " +
                type.toString());
            return null;
          }
        }
      }
      TypedIdentifier[] base0 = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
        base0[i] = canonicalTypedIdentifier(base[i]);
      Enumeration e = rtable.elements();
      L: while (e.hasMoreElements())
      {
        RecordType ctype = (RecordType)(e.nextElement());
        TypedIdentifier[] cbase = ctype.getBase();
        if (n != cbase.length) continue;
        for (int i=0; i<n; i++)
          if (base0[i] != cbase[i]) continue L;
        return ctype;
      }
      type = new RecordType(base0);
      rtable.add(type);
      return type;
    }
    
    if (type instanceof ArrayType)
    {
      ArrayType type0 = (ArrayType)type;
      Type index = canonical0(type0.getIndex());
      if (index == null) return null;
      Type base = canonical0(type0.getBase());
      if (base == null) return null;
      Enumeration e = atable.elements();
      while (e.hasMoreElements())
      {
        ArrayType ctype = (ArrayType)(e.nextElement());
        if (index == ctype.getIndex() && base == ctype.getBase())
          return ctype;
      }
      type = new ArrayType(index, base);
      atable.add(type);
      return type;
    }
    
    if (type instanceof SubType)
    {
      SubType stype = (SubType)type;
      Expression predicate = stype.getPredicate();
      Type ptype = process(predicate);
      if (ptype == null) return null;
      if (!(ptype instanceof FunctionType))
      {
        State.setError("subtype predicate " + predicate.toString() 
            + " has type " + ptype.toString() + 
            " which is not a function type");
        return type;
      }
      FunctionType ftype = (FunctionType)ptype;
      AST range = ftype.getRange();
      if (range != booleanType)
      {
        State.setError("subtype predicate " + predicate.toString() + 
            " has type " + ftype.toString() + 
            " whose range is not BOOL");
        return type;
      }
      // subtypes whose predicates refer to free variables
      // that are locally declared must not be shared
      Vector vars = FreeVariables.compute(predicate);
      boolean found = false;
      int n = vars.size();
      for (int i=0; i<n; i++)
      {
        ValueSymbol symbol = (ValueSymbol)vars.elementAt(i);
        if (symbol.getDepth() != 0)
        {
          found = true;
          break;
        }
      }
      if (!found)
      {
        Enumeration e = stable.elements();
        while (e.hasMoreElements())
        {
          SubType ctype = (SubType)(e.nextElement());
          if (predicate.toString().equals(ctype.getPredicate().toString()))
            return ctype;
        }
      }
      Type[] domain = ftype.getDomain();
      if (domain.length != 1)
      {
         State.setError("subtype predicate " + predicate.toString() + 
           " has type " + ftype.toString() + 
           " with " + domain.length + " parameters (1 expected)");
         return type;
      }
      stype.setBase(domain[0]);
      stable.add(stype);
      // type must not be empty
      Vector fv = FreeVariables.compute(predicate);
      int ext = varExtension(fv, "x", 0);
      ValueDeclIdentifier ident = Construct.valueDeclIdentifier("x"+ext);
      TypedIdentifier[] evars = 
        { Construct.typedIdentifier(ident , domain[0]) } ;
      Expression tccFormula =
        Construct.existsFormula(
          evars,
          Construct.applicationExpression(
            predicate, new Expression[] { Construct.reference(ident) }));
      process(tccFormula); // set reference value symbol
      Context context = State.getContext();
      context.implies(tccFormula);
      return stype;
    }
    
    if (type instanceof SubrangeType)
    {
      SubrangeType t = (SubrangeType)type;
      Expression from = t.getFrom();
      Expression to = t.getTo();
      Type ftype = process(from);
      if (ftype == null) return null;
      if (!hasType(from, ftype, intType, true))
      {
        State.setError("lower bound " + from.toString() +
          " of subrange type has type " + ftype.toString() +
          " (INT expected)");
        return null;
      }
      Type ttype = process(to);
      if (ttype == null) return null;
      if (!hasType(to, ttype, intType, true))
      {
        State.setError("upper bound " + to.toString() +
          " of subrange type has type " + ttype.toString() +
          " (INT expected)");
        return null;
      }
      // subrange types whose bounds refer to free variables
      // that are locally declared must not be shared
      Vector vars = FreeVariables.compute(from);
      boolean found = false;
      FreeVariables.append(to, vars);
      int n = vars.size();
      for (int i=0; i<n; i++)
      {
        ValueSymbol symbol = (ValueSymbol)vars.elementAt(i);
        if (symbol.getDepth() != 0) 
        {
          found = true;
          break;
        }
      }
      if (!found)
      {
        Enumeration e = utable.elements();
        while (e.hasMoreElements())
        {
          SubrangeType ctype = (SubrangeType)(e.nextElement());
          if (from.toString().equals(ctype.getFrom().toString()) && 
              to.toString().equals(ctype.getTo().toString()))
            return ctype;
        }
      }
      utable.add(t);
      // type must not be empty
      Context context = State.getContext();
      context.implies(Construct.lessEqualFormula(from, to));
      return t;
    }
    
    if (type instanceof LetType)
    {
      LetType t = (LetType)type;
      TypeDeclaration[] decls = t.getDeclarations();
      Type base = t.getBase();
      Environment env = State.getEnvironment();
      env.enter();
      Context context = State.getContext();
      int n = decls.length;
      for (int i=0; i<n; i++)
      {
        boolean okay = env.putTypeDeclaration(decls[i]) != null;
        if (!okay)
        {
          env.exit();
          for (int j=0; j<i; j++) context.abort();
          return null;
        }
        context.open();
      }
      Type base0 = canonical0(base);
      if (base0 == null)
      { for (int i=0; i<n; i++) context.abort(); }
      else
      { for (int i=0; i<n; i++) context.commit(); }
      context.close();
      env.exit();
      return base0;
    }
    
    State.setError("unknown type");
    return null;
  }

  /****************************************************************************
   * returns canonical version of typed identifier (equality of two types 
   * implies the equality of the canonical version of the types).
   * @param tid the typed identifier
   * @return the canonical version of the typed identifier
   **************************************************************************/
  private static TypedIdentifier canonicalTypedIdentifier(TypedIdentifier tid)
  {
    ValueDeclIdentifier id = tid.getIdentifier();
    String name = id.getName();
    Type base = canonical0(tid.getType());
    Enumeration e = titable.elements();
    while (e.hasMoreElements())
    {
      TypedIdentifier ctid = (TypedIdentifier)(e.nextElement());
      if (name.equals(ctid.getIdentifier().getName()) &&
          base == ctid.getType())
        return ctid;
    }
    tid = new TypedIdentifier(id, base);
    titable.add(tid);
    return tid;
  }
  
  /****************************************************************************
   * Returns canonical version of type (equality of two types implies
   * the equality of the canonical version of the types).
   * No type-checking conditions are generated.
   * @param type the type
   * @return the canonical version of the type (null, if an error occured)
   **************************************************************************/
  public static Type canonical(Type type)
  {
    Context context = State.getContext();
    context.open();
    Type t = canonical0(type);
    context.abort();
    return t;
  }
  
  /****************************************************************************
   * Returns canonical version of type (equality of two types implies
   * the equality of the canonical version of the types).
   * Type-checking conditions may be generaed.
   * @param type the type
   * @return the canonical version of the type (null, if an error occured)
   **************************************************************************/
  public static Type canonicalTcc(Type type)
  {
    return canonical0(type);
  }
  
  /****************************************************************************
   * get number from digit representation
   * @param digits the representation of the number as a decimal digit sequence
   * @return the number or null if conversion not possible
   ***************************************************************************/
  public static Integer parseInt(String digits)
  {
    try
    {
      return new Integer(Integer.parseInt(digits));
    }
    catch(NumberFormatException e)
    {
      State.setError("number " + digits + "cannot be represented as an int");
      return null;
    }
  }
  
  /****************************************************************************
   * type check value
   * @param value the value to be type checked
   * @return the type of the value or null if typechecking encountered error
   ***************************************************************************/
  public static Type process(Expression value)
  {
    // numbers ****************************************************************
    if (value instanceof fmrisc.Syntax.Number)
    {
      return natType;
    }
    
    // logical constant *******************************************************
    if (value instanceof Logical)
    {
      return booleanType;
    }
    
    // identifiers ************************************************************
    if (value instanceof Reference)
    {
      Reference ident = (Reference)value;
      ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
      if (symbol != null)
      {
        // already previously typechecked
        if (weakerTypes)
        {
          TypeExpression te = symbol.getClearType();
          if (te == null) 
            return symbol.getType();
          else
            return te.getType();
        }
        else
          return symbol.getType();
      }
      symbol = State.getEnvironment().getValueSymbol(ident);
      if (symbol == null)
      {
        State.setError("unknown identifier " + ident.toString());
        return null;
      }
      referencedSymbols.add(symbol);
      ident.setSymbol(symbol);
      return symbol.getType();
    }
    
    // unary expression ******************************************************
    if (value instanceof UnaryExpression)
    {
      Expression base = ((UnaryExpression)value).getBase();
      Type btype = process(base);
      if (btype == null) return null;
      if (value instanceof NegationTerm)
      {
        // no tcc in first call (real can be coerced to int by tcc)
        if (hasType(base, btype, intType, false)) return intType;
        if (hasType(base, btype, realType, true)) return realType;
        State.setError("expression " + base.toString() + 
            " has type " + btype.toString() +
            " (INT or REAL expected in arithmetic negation");
        return null;
      }
      if (value instanceof NotFormula)
      {
        if (hasType(base, btype, booleanType, true)) return booleanType;
        State.setError("expression " + base.toString() + 
            " has type " + btype.toString() +
            " (BOOL expected in logical negation)");
        return null;
      }
      State.setError("unknown kind of expression: " + value.toString());
      return null;
    }
    
    // binary expressions *****************************************************
    if (value instanceof BinaryExpression)
    {
      Expression first = ((BinaryExpression)value).getFirst();
      Expression second = ((BinaryExpression)value).getSecond();
      Type ftype = process(first);
      if (ftype == null) return null;        
      
      // second expression may be type checked in context of first one
      Context context = null;
      if (value instanceof OrFormula ||
          value instanceof AndFormula ||
          value instanceof ImpliesFormula)
      {
        context = State.getContext();
        context.open();
        if (value instanceof OrFormula)
          context.assume(Construct.notFormula(first));
        else
          context.assume(first);
      }
      Type stype = process(second);
      if (stype == null) 
      {
        if (context != null) context.abort();
        return null;
      }
      
      if (value instanceof EqualsFormula ||
          value instanceof NotEqualsFormula)
      {
        // try without tccs first
        //if (hasType(first, ftype, stype, false) ||
        //    hasType(second, stype, ftype, false) ||
        //    hasType(first, ftype, stype, true) ||
        //    hasType(second, stype, ftype, true)) 
        // no tcc can be generated by matching to BOOL, anyway
        if (hasType(first, ftype, booleanType, false))
        {
          State.setError("expression " + first.toString() + 
              " has type BOOLEAN " +
          " (rather than '=' or '/=' use '<=>' or 'XOR' on BOOLEAN values)");
          if (context != null) context.abort();
          return null;
        }
        if (superType(ftype, stype) != null)
        {
          if (context != null) context.commit();
          return booleanType;
        }
        State.setError("expression " + first.toString() +
            " has type " + ftype.toString() +
            " while expression " + second.toString() +
            " has type " + stype.toString() +
        " (matching types expected in (in)equality)");
        if (context != null) context.abort();
        return null;
      }
      
      if (value instanceof AndFormula ||
          value instanceof OrFormula ||
          value instanceof ImpliesFormula ||
          value instanceof EquivalentFormula ||
          value instanceof NotEquivalentFormula)
      {
        // no tcc can be generated by matching to BOOL, anyway
        if (!hasType(first, ftype, booleanType, false))
        {
          State.setError("expression " + first.toString() + 
              " has type " + ftype.toString() +
          " (BOOLEAN expected in logical connective)");
          if (context != null) context.abort();
          return null;
        }
        if (!hasType(second, stype, booleanType, true))
        {
          State.setError("expression " + second.toString() + 
              " has type " + stype.toString() +
          " (BOOLEAN expected in logical connective)");
          if (context != null) context.abort();
          return null;
        }  
        if (context != null) context.commit();
        return booleanType;
      }
      if (value instanceof PlusTerm ||
          value instanceof MinusTerm ||
          value instanceof TimesTerm ||
          value instanceof DividesTerm ||
          value instanceof PowerTerm ||
          value instanceof GreaterFormula ||
          value instanceof GreaterEqualFormula ||
          value instanceof LessFormula ||
          value instanceof LessEqualFormula)
      {
        // no tcc can be generated by matching to REAL, anyway
        if (!hasType(first, ftype, realType, false))
        {
          State.setError("expression " + first.toString() +
            " has type " + ftype.toString() +
            " (INT or REAL expected in arithmetic expression)");    
          if (context != null) context.abort();
          return null;
        }
        // generates TCC if value instanceof DividesTerm
        Type atype = null;
        if (value instanceof DividesTerm)
          atype = real0Type;
        else
          atype = realType;
        if (value instanceof PowerTerm)
        {
          if (!hasType(second, stype, intType, false))
          {
            State.setError("expression " + second.toString() +
              " has type " + stype.toString() +
              " (INT expected in power expression)");    
            if (context != null) context.abort();
            return null;
          }
        }
        else if (!hasType(second, stype, atype, true))
        {
          State.setError("expression " + second.toString() +
            " has type " + stype.toString() +
            " (INT or REAL expected in arithmetic expression)");    
          if (context != null) context.abort();
          return null;
        }   
        if (context != null) context.commit();
        if (value instanceof DividesTerm)
          return realType;
        else if (value instanceof PlusTerm ||
                 value instanceof MinusTerm ||
                 value instanceof TimesTerm)
        {
          if ((value instanceof PlusTerm ||
               value instanceof TimesTerm) &&
               hasType(first, ftype, natType, false) &&
               hasType(second, stype, natType, false))
            return natType;
          else if (hasType(first, ftype, intType, false) &&
                   hasType(second, stype, intType, false))
            return intType;
          else
            return realType;
        }
        else if (value instanceof PowerTerm)
        {
          if (hasType(first, ftype, intType, false))
          {
            if (hasType(second, stype, natType, false))
            {
              if (hasType(first, ftype, natType, false))
                return natType;
              else
                return intType;
            }
            else
              return realType;
          }
          else
            return realType;
        }
        else
          return booleanType;
      }
      if (context != null) context.abort();
      State.setError("unknown kind of expression: " + value.toString());
      return null;  
    }
       
    // conditional expressions ************************************************
    if (value instanceof IfThenElseExpression)
    {
      IfThenElseExpression exp = (IfThenElseExpression)value;
      Expression condition = exp.getCondition();
      Expression thenBranch = exp.getThenBranch();
      Expression elseBranch = exp.getElseBranch();
      Type ctype = process(condition);
      if (ctype == null) return null;
      // no tcc can be generated by matching to BOOL, anyway
      if (!hasType(condition, ctype, booleanType, false))
      {
        State.setError("Expression " + condition.toString() + 
            " has type " + ctype.toString() +
            " (BOOL expected in conditional expression) ");
        return null;
      }
      Context context = State.getContext();
      context.open();
      context.open();
      context.assume(condition);
      Type ttype = process(thenBranch);
      if (ttype == null)
      {
        context.abort();
        context.abort();
        return null;
      }
      context.commit();
      context.open();
      context.assume(Construct.notFormula(condition));
      Type etype = process(elseBranch);
      if (etype == null)
      {
        context.abort();
        context.abort();
        return null;
      }
      context.commit();
      // first try without generating TCC
      if (hasType(thenBranch, ttype, etype, false)) 
      {
        context.commit();
        return etype;
      }
      if (hasType(elseBranch, etype, ttype, false))
      {
        context.commit();
        return ttype;
      }
      Type stype = superType(ttype, etype);
      if (stype != null)
      {
        context.commit();
        return stype;
      }
      if (hasType(thenBranch, ttype, etype, true)) 
      {
        context.commit();
        return etype;
      }
      if (hasType(elseBranch, etype, ttype, true)) 
      { 
        context.commit();
        return ttype;
      }
      context.abort();
      State.setError("expression " + thenBranch.toString() +
          " has type " + ttype.toString() +
          " while expression " + elseBranch.toString() +
          " has type " + etype.toString() +
      " (equal types expected in conditional branches)");
      return null;
    }
    
    // let expressions *******************************************************
    if (value instanceof LetExpression)
    {
      LetExpression v = (LetExpression)value;
      ValueDeclaration[] decls = v.getDeclarations();
      Expression base = v.getBase();
      Environment env = State.getEnvironment();
      env.enter();
      Context context = State.getContext();
      int n = decls.length;
      for (int i=0; i<n; i++)
      {
        boolean okay = env.putValueDeclaration(decls[i]) != null;
        if (!okay)
        {
          env.exit();
          for (int j=0; j<i; j++) context.abort();
          return null;
        }
        context.open();
        ValueDeclIdentifier ident = decls[i].getValueDeclIdentifier();
        Expression dval = decls[i].getValue();
        Expression cond;
        if (isBoolean(ident.getSymbol().getType()))
          cond = Construct.equivalentFormula(Construct.reference(ident), dval);
        else
          cond = Construct.equalsFormula(Construct.reference(ident), dval);
        context.assume(cond);
      }
      Type type = process(base);
      if (type == null)
      { for (int i=0; i<n; i++) context.abort(); }
      else
      { for (int i=0; i<n; i++) context.commit(); }
      context.close();
      env.exit();
      return type;
    }
    
    // tuple terms ***********************************************************
    if (value instanceof TupleTerm)
    {
      TupleTerm v = (TupleTerm)value;
      Expression[] base = v.getBase();
      int n =  base.length;
      Type[] types = new Type[n];
      for (int i=0; i<n; i++)
      {
        Type t = process(base[i]);
        if (t == null) return null;
        types[i] = t;
      }
      return canonicalTcc(Construct.tupleType(types));
    }
    
    // record terms ***********************************************************
    if (value instanceof RecordTerm)
    {
      RecordTerm v = (RecordTerm)value;
      ValuedIdentifier[] base = v.getBase();
      int n =  base.length;
      TypedIdentifier[] types = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        Type t = process(base[i].getValue());
        if (t == null) return null;
        ValueDeclIdentifier ident = 
          Construct.valueDeclIdentifier(base[i].getIdentifier().getName());
        types[i] = Construct.typedIdentifier(ident, t);
      }
      return canonicalTcc(Construct.recordType(types));
    }
    
    // selection terms *******************************************************
    if (value instanceof SelectionTerm)
    {
      SelectionTerm v = (SelectionTerm)value; 
      Expression base = v.getBase();
      Selector sel = v.getSelector();
      Type btype = process(base);
      if (btype == null) return null;
      return getSelectedType(base, btype, sel);
    }
    
    // update term ***********************************************************
    if (value instanceof UpdateTerm)
    {
      UpdateTerm t = (UpdateTerm)value;
      Expression base = t.getBase();
      Selector[] sels = t.getSelectors();
      Expression update = t.getUpdate();
      Type btype = process(base);
      if (btype == null) return null;
      Expression exp = base;
      Type type = btype;
      int n = sels.length;
      for (int i=0; i<n; i++)
      {
        Selector sel = sels[i];
        type = getSelectedType(exp, type, sel);
        if (type == null) return null;
        exp = Construct.selectionTerm(exp, sel);
      }
      Type utype = process(update);
      if (utype == null) return null;
      if (hasType(update, utype, type, true)) return btype;
      State.setError(" type " + utype.toString() + 
          " of update value does not match type " + utype +
          " of " + exp);
      return null;
    }
    
    // function application ***************************************************
    if (value instanceof ApplicationExpression)
    {
      ApplicationExpression v = (ApplicationExpression)value;
      Expression fun = v.getFunction();
      Expression[] args = v.getArguments();
      Type ftype = process(fun);
      if (ftype == null) return null;
      if (!(ftype instanceof FunctionType))
      {
        State.setError("term " + fun.toString() +
            " has type " + ftype.toString() +
        " (function type expected in function application)");
        return null;
      }
      Type[] domain = ((FunctionType)ftype).getDomain();
      Type range = ((FunctionType)ftype).getRange();
      int n = domain.length;
      if (args.length != n)
      {
        State.setError("illegal application " + value.toString() +
          " of function with type " + ftype.toString() + 
          " to " + args.length + " arguments (" + n + 
          " arguments expected)");
        return null;
      }
      Context context = State.getContext();
      context.open();
      for (int i=0; i<n; i++)
      {
        Type atype = process(args[i]);
        if (atype == null) return null;
        if (hasType(args[i], atype, domain[i], true)) continue;
        State.setError("illegal application " + value.toString() + 
          " of function with type " + ftype.toString() +
          " to argument #" + i + " " + args[i].toString() +
          " of type " + atype.toString() + 
          " (argument of type " + domain[i].toString() + " expected)");
        context.abort();
        return null;
      }
      context.commit();
      return range;
    }
    
    // quantified expressions *************************************************
    if (value instanceof QuantifiedExpression)
    {
      QuantifiedExpression v = (QuantifiedExpression)value;
      TypedIdentifier[] vars = v.getVariables();
      Expression base = v.getBase();
      Environment env = State.getEnvironment();
      env.enter();
      boolean okay = env.putVariables(vars);
      if (!okay) 
      {
        env.exit();
        return null;
      }
      Type type = process(base);
      if (type == null) 
      {
        env.exit();
        return null;
      }
      Context context = State.getContext();
      if (value instanceof ForallFormula || value instanceof ExistsFormula)
      {
        okay = hasType(base, type, booleanType, true);
        context.close();
        env.exit();
        simplifyNames(vars, env, base);
        if (okay) return booleanType;
        State.setError("body of quantified formula has type " + 
            type.toString() + 
            " (BOOL expected)");
        return null;
      }
      context.close();
      env.exit();
      simplifyNames(vars, env, base);
      int n = vars.length;
      Type[] ptypes = new Type[n];
      for (int i=0; i<n; i++)
      {
        ptypes[i] = canonicalTcc(vars[i].getType());
      }
      if (value instanceof LambdaTerm)
        return canonicalTcc(Construct.functionType(ptypes, type));
      else if (value instanceof ArrayTerm)
      {
        if (n != 1)
        {
          State.setError("ARRAY expression " + value.toString() +
            " has " + n + " parameters (1 expected)");
          return null;
        }
        return canonicalTcc(Construct.arrayType(ptypes[0], type));
      }
      State.setError("unknown kind of quantiifier: " + value.toString());
      return null;
    }
    
    State.setError("unknown kind of expression: " + value.toString());
    return null;
  }

  /****************************************************************************
   * returns type of selected expression
   * @param base the base expression
   * @param btype the type of the base expression
   * @param sel the selector
   * @return the selected type or null, if something went wrong
   ***************************************************************************/
  private static Type getSelectedType(Expression base, 
    Type btype, Selector sel)
  {
    // first strip subtypes from btype
    while (btype instanceof SubType)
    {
      SubType stype = (SubType)btype;
      btype = stype.getBase();
    }
    if (sel instanceof SelectorNumber)
    {
      if (!(btype instanceof TupleType))
      {
        State.setError("value " + base.toString() +
            " has type " + btype.toString() +
        " (TUPLE type expected for component selection by number)");
        return null;
      }
      Type[] btypes = ((TupleType)btype).getBase();
      Integer number = parseInt(
          ((SelectorNumber)sel).getNumber().getDigits());
      if (number == null) return null;
      int n = number.intValue();
      int l = btypes.length;
      if (n < 0 || l <= n)
      {
        State.setError("type " + btype.toString() +
            " of term " + base.toString() + 
            " only allows selection of components 0.." + (n-1) +
            " (component " + n + " selected)");
        return null;
      }
      return btypes[n];
    }
    if (sel instanceof SelectorIdentifier)
    {     
      if (!(btype instanceof RecordType))
      {
        State.setError("value " + base.toString() +
            " has type " + btype.toString() +
        " (RECORD type expected for component selection by identifier)");
        return null;
      }
      String name = ((SelectorIdentifier)sel).getIdentifier().getName();
      TypedIdentifier[] tids = ((RecordType)btype).getBase();
      int l = tids.length;
      for (int i=0; i<l; i++)
      {
        TypedIdentifier tid = tids[i];
        if (name.equals(tid.getIdentifier().getName()))
          return tid.getType();
      }
      State.setError("type " + btype.toString() +
          " of term " + base.toString() + 
          " does not have a component called " + name);
      return null;
    }
    if (sel instanceof SelectorIndex)
    {
      if (!(btype instanceof ArrayType))
      {
        State.setError("value " + base.toString() +
            " has type " + btype.toString() +
        " (ARRAY type expected for component selection by index)");
        return null;
      }
      Type itype = ((ArrayType)btype).getIndex();
      Type etype = ((ArrayType)btype).getBase();
      Expression index = ((SelectorIndex)sel).getIndex();
      Type stype = process(index);
      if (stype == null) return null;
      if (!hasType(index, stype, itype, true))
      {
        State.setError("index " + index.toString() +
            " has type " + stype.toString() +
            " which does not match index type " + itype.toString() +
            " of array " + base.toString());
        return null;
      }
      return etype;
    }
    State.setError("unknown kind of selector: " + sel.toString());
    return null;
  }
  
  /***************************************************************************
   * returns true iff expr of type etype matches type.
   * may add as a side effect a type checking condition to the global context.
   * 
   * the following general rules succeed without generating TCCs:
   *    
   *  exp: T                  matches T
   *  exp: SUBTYPE(p:T->BOOL) matches T
   * 
   * the following general rule succeeds by generating a TCC:
   * 
   *   exp: T      matches SUBTYPE(T)         if p(exp)
   *    
   * the following structural rules do not generate TCCs:
   * 
   *   exp: A->B         matches A0->B0,         if A0 matches A and B matches B0
   *   exp: ARRAY A OF B matches ARRAY A0 OF B0, if A0 matches A and B matches B0
   * 
   * the following structural rules *may* generate TCCs:
   * 
   *   exp: [T,...]      matches [T0,...],       if exp.0: T matches T0, ...
   *   exp: [#f:T,...#]  matches [#f:T0,...],    if exp.f: T matches T0, ...
   * 
   * as for arithmetic, the relationship is as follows:
   *
   *   [a..b] < INT   < REAL
   *   NAT    < INT
   *            REAL0 < REAL
   *   
   * (REAL0, i.e. REAL without 0, is only internally used to denote
   * the second argument type of the / operator and thus trigger a TCC)
   *
   * the corresponding conversions from left to right succeed without
   * generating TCCs; conversions are applied transitively.
   * 
   * as for the conversion from right to left, we have:
   * 
   *   exp: REAL   matches INT      if (EXISTS (x:INT): x=exp)
   *   // we switch this rule off in order to not confuse CVCL
   *   
   *   exp: INT    matches [a..b]   if a <= exp && exp <= b
   *   exp: INT    matches NAT      if 0 <= exp
   *   exp: REAL   matches REAL0    if exp /= 0
   *    
   * as for subranges of INT, we have
   * 
   *   exp: NAT matches [a..b]      if (a <= 0 || a <= exp) && (exp <= b)
   *   exp: [a..b] matches NAT      if 0 <= a || 0 <= exp
   *   exp: [a..b] matches [c..d]   if (c <= a || c <= exp) &&
   *                                   (b <= d || exp <= d)
   *                                                                 
   * if rules are applied transitively, the conditions are appropriately
   * conjoined, e.g. exp: REAL -> SUBTYPE(p:[a..b]->BOOL) generates the TCC
   * 
   *   (EXISTS (x:INT): x=exp) && a <= exp && exp <= b && p(exp)
   * 
   * @param exp the expression (may be null if tcc is false)
   * @param etype its type
   * @param type the type which expr:etype is expected to match
   * @param tcc true iff a tcc may be generated
   * @return true iff the matching succeeds
   **************************************************************************/
  public static boolean hasType(Expression exp, Type etype, Type type,
    boolean tcc)
  {
    // exp: T matches T
    if (etype == type) return true;
    
    // exp: SUBTYPE(p:T->BOOL) matches T (with assumption p(exp))
    if (etype instanceof SubType)
    {
      SubType stype = (SubType)etype;
      Type domain = stype.getBase();
      Context context = null;
      if (tcc)
      {
        context = State.getContext();
        context.open();
        Expression[] args = { exp };
        context.assume(Construct.applicationExpression(
          stype.getPredicate(), args));
      }
      boolean result = hasType(exp, domain, type, tcc);
      if (tcc) 
      {
        if (result)
          context.commit();
        else
          context.abort();
      }
      return result;
    }
    
    // exp: T matches SUBTYPE(T) if p(exp)
    // in non-generating type mode, we just check for base type compatibility
    if (type instanceof SubType)
    {
      SubType stype = (SubType)type;
      Type domain = stype.getBase();
      Context context = State.getContext();
      context.open();
      if (!hasType(exp, etype, domain, tcc))
      {
        context.abort();
        return false;
      }
      if (!tcc) 
      {
        context.abort();
        return true;
      }
      Expression[] args = { exp };
      Expression tccFormula = 
        Construct.applicationExpression(stype.getPredicate(), args);
      context.implies(tccFormula);
      context.commit();
      return true;
    }
    
    // exp: A->B matches A0->B0, if A0 matches A and B matches B0
    if (etype instanceof FunctionType && type instanceof FunctionType)
    {
      FunctionType etype0 = (FunctionType)etype;
      FunctionType type0 = (FunctionType)type;
      
      // matching of domain
      Type[] domain = type0.getDomain();
      Type[] edomain = etype0.getDomain();
      int n = domain.length;
      if (n != edomain.length) return false;
      for (int i=0; i<n; i++)
      {
        if (strongerTypes)
        {
          // also covariant matching of domain
          if (!hasType(null, edomain[i], domain[i], false)) return false;
        }
        // contravariant matching of domain
        if (!hasType(null, domain[i], edomain[i], false)) return false;  
      }
      
      if (strongerTypes)
      {
        // also contravariant matching of range
        if (!hasType(null, type0.getRange(), etype0.getRange(), false)) 
          return false;
      }
      // covariant matching of range
      return hasType(null, etype0.getRange(), type0.getRange(), false);
    }
   
    // exp: ARRAY A OF B matches ARRAY A0 OF B0, if A0 matches A and B matches B0
    if (etype instanceof ArrayType && type instanceof ArrayType)
    {
      ArrayType etype0 = (ArrayType)etype;
      ArrayType type0 = (ArrayType)type;
      
      // switch off covariant/contravariant matching
      if (strongerTypes) 
      {
        if (!hasType(null, etype0.getIndex(), type0.getIndex(), false) ||
            !hasType(null, type0.getBase(), etype0.getBase(), false))
          return false;
      }
      
      // array type is contravariant in index, no TCCs generated
      return (hasType(null, type0.getIndex(), etype0.getIndex(), false) &&
              hasType(null, etype0.getBase(), type0.getBase(), false));
    }
    
    // exp: [T,...] matches [T0,...], if exp.0: T matches T0, ...
    if (etype instanceof TupleType && type instanceof TupleType)
    {
      TupleType etype0 = (TupleType)etype;
      TupleType type0 = (TupleType)type;
      Type[] etype1 = etype0.getBase();
      Type[] type1 = type0.getBase();
      int n = etype1.length;
      if (n != type1.length) return false;
      Context context = null;
      if (tcc) 
      {
        context = State.getContext();
        context.open();
      }
      for (int i=0; i<n; i++)
      {
        Expression exp0 = null;
        if (tcc)
          exp0 = Construct.selectionTerm(exp,
            Construct.selectorNumber(
              Construct.number(Integer.toString(i))));
        if (!hasType(exp0, etype1[i], type1[i], tcc)) 
        {
          if (tcc) context.abort();
          return false;  
        }
      }
      if (tcc) context.commit();
      return true;
    }
    
    // exp: [#f:T,...#]  matches [#f:T0,...],    if exp.f: T matches T0, ...
    if (etype instanceof RecordType && type instanceof RecordType)
    {
      RecordType etype0 = (RecordType)etype;
      RecordType type0 = (RecordType)type;
      TypedIdentifier[] etype1 = etype0.getBase();
      TypedIdentifier[] type1 = type0.getBase();
      int n = etype1.length;
      if (n != type1.length) return false;
      Context context = null;
      if (tcc) 
      {
        context = State.getContext();
        context.open();
      }
      for (int i=0; i<n; i++)
      {
        ValueDeclIdentifier ident = etype1[i].getIdentifier();
        if (!ident.getName().equals(type1[i].getIdentifier().getName())) 
        {
          if (tcc) context.abort();
          return false;
        }
        Expression exp0 = null;
        if (tcc)
        {
          Reference ref = Construct.reference(ident.getName());
          ref.setSymbol(ident.getSymbol());
          exp0 = Construct.selectionTerm(exp,
            Construct.selectorIdentifier(ref));
        }
        if (!hasType(exp0, etype1[i].getType(), type1[i].getType(), tcc))
        {
          if (tcc) context.abort();
          return false;
        }
      }
      if (tcc) context.commit();
      return true;
    }
    
    // arithmetic conversions without tccs
    if (etype instanceof SubrangeType && (type == intType || type == realType)) 
      return true;
    if (etype == natType && (type == intType || type == realType))
      return true;
    if (etype == intType && type == realType)
      return true;
    
    // weaker types do not have NAT, only INT
    if (weakerTypes && etype == intType && type == natType)
      return true;
    
    // after that, only chance is to generate tccs
    if (!tcc) return false;
      
    // [a..b]->REAL0               exp /= 0
    // NAT->REAL0                  exp /= 0
    // INT->REAL0                  exp /= 0
    // REAL->REAL0                 exp /= 0
    if (type == real0Type)
    {
      if (!(etype instanceof SubrangeType) && etype != natType &&
          etype != intType && etype != realType)
        return false;
      Expression tccFormula =
        Construct.notEqualsFormula(exp, Construct.number("0"));
      Context context = State.getContext();
      context.implies(tccFormula);
      return true;
    }

    // INT->[c..d]      c <= exp && exp <= d
    // NAT->[c..d]      (c <= 0 || c <= exp) && exp <= d
    // [a..b]->[c..d]   (c <= a || c <= exp) && (b <= d || exp <= d)   
    if ((etype == intType || etype == natType || etype instanceof SubrangeType) && 
        type instanceof SubrangeType)
    {
      SubrangeType stype = (SubrangeType)type;
      Expression from = stype.getFrom();
      Expression to = stype.getTo();
      Expression tcc1 = Construct.lessEqualFormula(from, exp);
      Expression tcc2 = Construct.lessEqualFormula(exp, to);
      if (etype == natType)
        tcc1 = Construct.orFormula(
         Construct.lessEqualFormula(from, Construct.number("0")),
         tcc1);
      else if (etype instanceof SubrangeType)
      {
        SubrangeType etype0 = (SubrangeType)etype;
        tcc1 = Construct.orFormula(
            Construct.lessEqualFormula(from, etype0.getFrom()),
            tcc1);
        tcc2 = Construct.orFormula(
            Construct.lessEqualFormula(etype0.getTo(), to),
            tcc2);
      }
      Expression tccFormula = Construct.andFormula(tcc1, tcc2);
      Context context = State.getContext();
      context.implies(tccFormula);
      return true;
    }
    
    // INT->NAT         0 <= exp
    if (etype == intType && type == natType)
    {
      Expression tccFormula =
          Construct.greaterEqualFormula(exp, Construct.number("0"));
      Context context = State.getContext();
      context.implies(tccFormula);
      return true;
    }
    
    // [a..b]->NAT         (0 <= a) || (0 <= exp)
    if (etype instanceof SubrangeType && type == natType)
    {
      Expression from = ((SubrangeType)etype).getFrom();
      Expression tccFormula =
          Construct.orFormula(
            Construct.greaterEqualFormula(from, Construct.number("0")),
            Construct.greaterEqualFormula(exp, Construct.number("0"))
          );
      Context context = State.getContext();
      context.implies(tccFormula);
      return true;
    }
    
    // we switch this rule off in order to not confuse CVCL
    /*

    // REAL->INT        EXISTS(x:INT):x=exp
    if (false && etype == realType && type == intType)
    {
      Vector fv = FreeVariables.compute(exp);
      int ext = varExtension(fv, "x", 0);
      ValueDeclIdentifier ident = Construct.valueDeclIdentifier("x"+ext);
      TypedIdentifier[] vars = 
        { Construct.typedIdentifier(
            ident ,
            Construct.atomicType("INT")) } ;
      Expression tccFormula =
        Construct.existsFormula(
          vars,
          Construct.equalsFormula(Construct.reference(ident), exp));
      process(tccFormula); // set reference value symbol
      Context context = State.getContext();
      context.implies(tccFormula);
      return true;
    }
    
    // REAL->NAT        EXISTS(x:NAT):x=exp
    if (etype == realType && type == natType)
    {
      Vector fv = FreeVariables.compute(exp);
      int ext = varExtension(fv, "x", 0);
      ValueDeclIdentifier ident = Construct.valueDeclIdentifier("x"+ext);
      TypedIdentifier[] vars = 
        { Construct.typedIdentifier(
            ident ,
            Construct.atomicType("NAT")) } ;
      Expression tccFormula =
        Construct.existsFormula(
          vars,
          Construct.equalsFormula(Construct.reference(ident), exp));
      process(tccFormula); // set reference value symbol
      Context context = State.getContext();
      context.implies(tccFormula);
      return true;
    }
    
    // REAL->[a..b]     (EXISTS(x:INT):x=exp && a<=x && x<=b)
    if (etype == realType && type instanceof SubrangeType)
    {
      SubrangeType stype = (SubrangeType)type;
      Expression from = stype.getFrom();
      Expression to = stype.getTo();
      Vector fv = FreeVariables.compute(exp);
      FreeVariables.append(from, fv);
      FreeVariables.append(to, fv);
      int ext = varExtension(fv, "x", 0);
      ValueDeclIdentifier ident = Construct.valueDeclIdentifier("x"+ext);
      TypedIdentifier[] vars = 
      { Construct.typedIdentifier(
          ident,
          Construct.atomicType("INT")) } ;
      Expression tccFormula =
        Construct.existsFormula(
          vars,
          Construct.andFormula(
            Construct.andFormula(
              Construct.equalsFormula(Construct.reference(ident), exp),
              Construct.lessEqualFormula(from, Construct.reference(ident))),
            Construct.lessEqualFormula(Construct.reference(ident), to)));
      process(tccFormula); // set reference value symbol
      Context context = State.getContext();
      context.implies(tccFormula);
      return true;   
    }
    */
    
    // bad luck
    return false;
  }
  
  /***************************************************************************
   * return smallest supertype of type0 and type1 (null, if none exists)
   * does not generate tccs.
   * @param type0 a type
   * @param type1 a type
   * @return the supertype of type0 and type1 (or null, if none exists)
   *************************************************************************/
  private static Type superType(Type type0, Type type1)
  {
    if (type0 == type1) return type0;
    
    if (type0 instanceof SubType)
    {
      SubType stype = (SubType)type0;
      return superType(stype.getBase(), type1);
    }
    if (type1 instanceof SubType)
    {
      SubType stype = (SubType)type1;
      return superType(type0, stype.getBase());
    }
    
    if ((type0 instanceof SubrangeType || type0 == natType) &&
         (type1 == intType || type1 == realType)) return type1;
    if ((type1 instanceof SubrangeType || type1 == natType) &&
        (type0 == intType || type0 == realType)) return type0;
    
    if ((type0 instanceof SubrangeType || type0 == natType) && 
        (type1 instanceof SubrangeType || type1 == natType))
      return intType;
    
    if (type0 == intType && type1 == realType) return type1;
    if (type1 == intType && type0 == realType) return type0;
    
    if ((type0 instanceof SubrangeType || type0 == natType || type0 == intType)
          && type1 == real0Type) return realType;
    if ((type1 instanceof SubrangeType || type1 == natType || type1 == intType) 
          && type0 == real0Type) return realType;
    
    if (type0 == real0Type && type1 == realType) return type1;
    if (type1 == real0Type && type0 == realType) return type0;
    
    if (type0 instanceof FunctionType && type1 instanceof FunctionType)
    {
      FunctionType ftype0 = (FunctionType)type0;
      FunctionType ftype1 = (FunctionType)type1;
      Type[] domain0 = ftype0.getDomain();
      Type range0 = ftype0.getRange();
      Type[] domain1 = ftype1.getDomain();
      Type range1 =ftype1.getRange();
      int n = domain0.length;
      if (n != domain1.length) return null;
      Type[] domain = new Type[n];
      for (int i=0; i<n; i++)
      {
        domain[i] = subType(domain0[i], domain1[i]);
        if (domain[i] == null) return null;
      }
      Type range = superType(range0, range1);
      if (range == null) return null;
      return canonical(Construct.functionType(domain, range));
    }
    
    if (type0 instanceof ArrayType && type1 instanceof ArrayType)
    {
      ArrayType ftype0 = (ArrayType)type0;
      ArrayType ftype1 = (ArrayType)type1;
      Type domain0 = ftype0.getIndex();
      Type range0 = ftype0.getBase();
      Type domain1 = ftype1.getIndex();
      Type range1 =ftype1.getBase();
      Type domain = subType(domain0, domain1);
      if (domain == null) return null;
      Type range = superType(range0, range1);
      if (range == null) return null;
      return canonical(Construct.arrayType(domain, range));
    }
    
    if (type0 instanceof TupleType && type1 instanceof TupleType)
    {
      Type[] base0 = ((TupleType)type0).getBase();
      Type[] base1 = ((TupleType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      Type[] base = new Type[n];
      for (int i=0; i<n; i++)
      {
        base[i] = superType(base0[i], base1[i]);
        if (base[i] == null) return null;
      }
      return canonical(Construct.tupleType(base));
    }
    
    if (type0 instanceof RecordType && type1 instanceof RecordType)
    {
      TypedIdentifier[] base0 = ((RecordType)type0).getBase();
      TypedIdentifier[] base1 = ((RecordType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      TypedIdentifier[] base = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        String name = base0[i].getIdentifier().getName();
        if (!name.equals(base1[i].getIdentifier().getName()))
          return null;
        Type type = superType(base0[i].getType(), base1[i].getType());
        if (type == null) return null;
        base[i] = Construct.typedIdentifier(
          Construct.valueDeclIdentifier(name), type);
      }
      return canonical(Construct.recordType(base));
    }
    
    return null;
  }
  
  /***************************************************************************
   * return greatest subtype of type0 and type1 (null, if none exists)
   * does not generate tccs.
   * @param type0 a type
   * @param type1 a type
   * @return the subtype of type0 and type1 (or null, if none exists)
   *************************************************************************/
  public static Type subType(Type type0, Type type1)
  {
    if (type0 == type1) return type0;
    
    if (type0 instanceof SubType && hasType(null, type0, type1, false))
      return type0;
    if (type1 instanceof SubType && hasType(null, type1, type0, false))
      return type1;
    
    if ((type0 instanceof SubrangeType || type0 == natType) &&
         (type1 == intType || type1 == realType)) return type0;
    if ((type1 instanceof SubrangeType || type1 == natType) &&
        (type0 == intType || type0 == realType)) return type1;
    
    if (type0 == intType && type1 == realType) return type0;
    if (type1 == intType && type0 == realType) return type1;
    
    if (type0 == real0Type && type1 == realType) return type0;
    if (type1 == real0Type && type0 == realType) return type1;
    
    if (type0 instanceof FunctionType && type1 instanceof FunctionType)
    {
      FunctionType ftype0 = (FunctionType)type0;
      FunctionType ftype1 = (FunctionType)type1;
      Type[] domain0 = ftype0.getDomain();
      Type range0 = ftype0.getRange();
      Type[] domain1 = ftype1.getDomain();
      Type range1 =ftype1.getRange();
      int n = domain0.length;
      if (n != domain1.length) return null;
      Type[] domain = new Type[n];
      for (int i=0; i<n; i++)
      {
        domain[i] = superType(domain0[i], domain1[i]);
        if (domain[i] == null) return null;
      }
      Type range = subType(range0, range1);
      if (range == null) return null;
      return canonical(Construct.functionType(domain, range));
    }
    
    if (type0 instanceof ArrayType && type1 instanceof ArrayType)
    {
      ArrayType ftype0 = (ArrayType)type0;
      ArrayType ftype1 = (ArrayType)type1;
      Type domain0 = ftype0.getIndex();
      Type range0 = ftype0.getBase();
      Type domain1 = ftype1.getIndex();
      Type range1 =ftype1.getBase();
      Type domain = superType(domain0, domain1);
      if (domain == null) return null;
      Type range = subType(range0, range1);
      if (range == null) return null;
      return canonical(Construct.arrayType(domain, range));
    }
    
    if (type0 instanceof TupleType && type1 instanceof TupleType)
    {
      Type[] base0 = ((TupleType)type0).getBase();
      Type[] base1 = ((TupleType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      Type[] base = new Type[n];
      for (int i=0; i<n; i++)
      {
        base[i] = subType(base0[i], base1[i]);
        if (base[i] == null) return null;
      }
      return canonical(Construct.tupleType(base));
    }
    
    if (type0 instanceof RecordType && type1 instanceof RecordType)
    {
      TypedIdentifier[] base0 = ((RecordType)type0).getBase();
      TypedIdentifier[] base1 = ((RecordType)type1).getBase();
      int n = base0.length;
      if (n != base1.length) return null;
      TypedIdentifier[] base = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        String name = base0[i].getIdentifier().getName();
        if (!name.equals(base1[i].getIdentifier().getName()))
          return null;
        Type type = subType(base0[i].getType(), base1[i].getType());
        if (type == null) return null;
        base[i] = Construct.typedIdentifier(
          Construct.valueDeclIdentifier(name), type);
      }
      return canonical(Construct.recordType(base));
    }
    
    return null;
  }
  
  /***************************************************************************
   * return version of type where subtypes and subranges (including NAT) are 
   * replaced by their base types; the corresponding constraints are 
   * returned as a formula which a corresponding expression exp has to fulfill
   * @param type the type to be cleared
   * @param exp an expression of this type
   * @return a pair of cleared type and
   *         formula (may be null if no condition generated)
   *         or null (if cleared type is identical to given type)
   **************************************************************************/
  public static TypeExpression clearSubtypes(Type type, Expression exp)
  { 
    if (type == natType)
    {
      Expression formula =
        Construct.greaterEqualFormula(exp, Construct.number("0"));
      return new TypeExpression(intType, formula.instantiate());
    }
    
    if (type instanceof AtomicType)
      return null;
    
    if (type instanceof SubrangeType)
    {
      SubrangeType type0 = (SubrangeType)type;
      Expression formula = Construct.andFormula(
        Construct.lessEqualFormula(type0.getFrom(), exp),
        Construct.lessEqualFormula(exp, type0.getTo()));
      return new TypeExpression(intType, formula.instantiate());
    }
    
    if (type instanceof SubType)
    {
      SubType type0= (SubType)type;
      Type base = type0.getBase();
      TypeExpression te = clearSubtypes(base, exp);
      if (te != null) base = te.getType();
      Expression[] args = { exp };
      Expression formula = 
        Construct.applicationExpression(type0.getPredicate(), args);
      if (te != null && te.getExpression() != null)
        formula = Construct.andFormula(te.getExpression(), formula);
      return new TypeExpression(base, formula.instantiate());
    }
    
    if (type instanceof TupleType)
    {
      TupleType type0 = (TupleType)type;
      Type[] base = type0.getBase();
      int n = base.length;
      Type[] base0 = new Type[n];
      Expression formula = null;
      boolean changed = false;
      for (int i=0; i<n; i++)
      {
        Expression exp0 =
          Construct.selectionTerm(exp,
            Construct.selectorNumber(
              Construct.number(Integer.toString(i))));
        TypeExpression te = clearSubtypes(base[i], exp0);
        if (te == null)
        {
          base0[i] = base[i];
          continue;
        }
        changed = true;
        base0[i] = te.getType();  
        if (formula == null)
          formula = te.getExpression();
        else if (te.getExpression() != null)
          formula = Construct.andFormula(formula, te.getExpression());
      }
      if (!changed) return null;
      return new TypeExpression(
        canonical(Construct.tupleType(base0)),
        formula); 
    }
    
    if (type instanceof RecordType)
    {
      RecordType type0 = (RecordType)type;
      TypedIdentifier[] base = type0.getBase();
      int n = base.length;
      TypedIdentifier[] base0 = new TypedIdentifier[n];
      Expression formula = null;
      boolean changed = false;
      for (int i=0; i<n; i++)
      {
        ValueDeclIdentifier ident = base[i].getIdentifier();
        Expression exp0 =
          Construct.selectionTerm(exp,
            Construct.selectorIdentifier(Construct.reference(ident)));
        TypeExpression te = clearSubtypes(base[i].getType(), exp0);
        if (te == null)
        {
          base0[i] = base[i];
          continue;
        }
        changed = true;
        base0[i] = Construct.typedIdentifier(ident, te.getType());  
        if (formula == null)
          formula = te.getExpression();
        else if (te.getExpression() != null)
          formula = Construct.andFormula(formula, te.getExpression());
      }
      if (!changed) return null;
      return new TypeExpression(
        canonical(Construct.recordType(base0)),
        formula); 
    }
    
    if (type instanceof FunctionType)
    {
      FunctionType type0 = (FunctionType)type;
      Type[] domain = type0.getDomain();
      Type range = type0.getRange();
      int n = domain.length;
      ValueDeclIdentifier[] ident = new ValueDeclIdentifier[n];
      Vector fv = FreeVariables.compute(type0);
      FreeVariables.append(exp, fv);
      int ext = 0;
      for (int i=0; i<n; i++)
      {
        int ext0 = varExtension(fv, "x", ext);
        ident[i] = new ValueDeclIdentifier("x"+ext0);
        ext = ext0+1;
      }
      TypeExpression[] arg = new TypeExpression[n];
      Type[] atypes = new Type[n];
      boolean argnull = true;
      for (int i=0; i<n; i++)
      {
        arg[i] = clearSubtypes(domain[i], Construct.reference(ident[i]));
        if (arg[i] == null)
          atypes[i] = domain[i];
        else
        {
          atypes[i] = arg[i].getType();
          argnull = false;
        }
      }
      Expression[] args = new Expression[n];
      for (int i=0; i<n; i++)
        args[i] = Construct.reference(ident[i]);
      TypeExpression result =
        clearSubtypes(range, Construct.applicationExpression(exp, args));
      if (result == null)
      {
        if (argnull)
          return null;
        else
          return new TypeExpression(
            canonical(Construct.functionType(atypes, range)),
            null); // no condition on result required
      }
      TypedIdentifier[] tident = new TypedIdentifier[n];
      for (int i=0; i<n; i++)
      {
        tident[i] = Construct.typedIdentifier(ident[i], atypes[i]);
        ident[i].getSymbol().setType(atypes[i]);
      }
      Expression formula = result.getExpression();
      if (formula != null)
      {
        Expression assumption = null;
        for (int i=0; i<n; i++)
        {
          if (arg[i] != null && arg[i].getExpression() != null)
          {
            if (assumption == null)
              assumption = arg[i].getExpression();
            else
              assumption = Construct.andFormula(assumption, arg[i].getExpression());
          }
        }
        if (assumption != null)
          formula = Construct.impliesFormula(assumption, formula);
        formula = Construct.forallFormula(tident, formula);
        formula = formula.instantiate();
      }
      return new TypeExpression(
        canonical(Construct.functionType(atypes, result.getType())), formula);
    }
    
    if (type instanceof ArrayType)
    {
      ArrayType type0 = (ArrayType)type;
      Type domain = type0.getIndex();
      Type range = type0.getBase();
      Vector fv = FreeVariables.compute(type);
      FreeVariables.append(exp, fv);
      int ext = varExtension(fv, "x", 0);
      ValueDeclIdentifier ident = new ValueDeclIdentifier("x"+ext);
      TypeExpression arg = 
        clearSubtypes(domain, Construct.reference(ident));
      Selector index =
        Construct.selectorIndex(Construct.reference(ident));
      TypeExpression result =
        clearSubtypes(range, Construct.selectionTerm(exp, index));
      if (result == null)
      {
        if (arg == null)
          return null;
        else
          return new TypeExpression(
            canonical(Construct.arrayType(arg.getType(), range)),
            null); // no condition on result required
      }
      if (arg != null) domain = arg.getType();
      TypedIdentifier tident = Construct.typedIdentifier(ident, domain);
      ident.getSymbol().setType(domain);
      Expression formula = result.getExpression();
      if (arg != null && arg.getExpression() != null && formula != null)
        formula = Construct.impliesFormula(arg.getExpression(), formula);
      if (formula != null)
      {
        TypedIdentifier[] tidents = { tident };
        formula = Construct.forallFormula(tidents, formula);
        formula = formula.instantiate();
      }
      return new TypeExpression(
        canonical(Construct.arrayType(domain, result.getType())),
        formula);
    }
    
    return null;
  }

  /****************************************************************************
   * Decides whether name occurs among vector of ValueSymbol objects.
   * @param fv a vector of ValueSymbol objects.
   * @param name a name
   * @return true if some of the objects in fv is named name.
   ***************************************************************************/
  private static boolean contains(Vector fv, String name)
  {
    int n = fv.size();
    for (int i=0; i<n; i++)
    {
      ValueSymbol symbol = (ValueSymbol)fv.elementAt(i);
      if (name.equals(symbol.getIdentifier().getName())) return true;
    }
    return false;
  }
    
  /***************************************************************************
   * Create a new variable name.
   * @param fv a vector of ValueSymbol objects.
   * @param base the base of the new name.
   * @param ext the start of a range of numeric extensions.
   * @return a numeric extension that if added to the base yiels a name that
   *         does not collide with the names of the value symbols.
   **************************************************************************/
  private static int varExtension(Vector fv, String base, int ext)
  {
    while(true)
    {
      if (!contains(fv, base+ext)) return ext;
      ext = ext+1;
    }
  }
  
  /***************************************************************************
   * Simplify names of local variables to their base names 
   * if no name capture can occur.
   * @param vars local variables
   * @param env the environment in which the local variables are introduced.
   * @param exp the expression within which the variables are referenced.
   *************************************************************************/
  public static void simplifyNames(TypedIdentifier[] vars, Environment env,
    Expression exp)
  {
    Vector bound = BoundVariables.compute(exp);
    int n = vars.length;
    loop: for (int i=0; i<n; i++)
    {
      ValueDeclIdentifier ident = vars[i].getIdentifier();
      String base = ASTUtil.getBaseName(ident.getName());
      if (env.getValueSymbol(Construct.reference(base)) == null &&
          !bound.contains(base))
      {
        for (int j=0; j<i; j++)
        {
          // name must also not be captured by name of previous variable 
          if (base.equals(vars[j].getIdentifier().getName()))
            continue loop;
        }
        ident.setName(base);
      }
    }
  }
}
