// ----------------------------------------------------------------------------
// TypedIdentifier.java
// Handling of typed identifiers.
// $Id: TypedIdentifier.java,v 1.12 2006/02/23 19:51:07 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Syntax;

import java.io.PrintWriter;

/**
 * Handling of typed identifiers.
 */
public final class TypedIdentifier extends ASTBase
{
  private ValueDeclIdentifier identifier; // the identifier
  private Type type;                      // its type
  
  /****************************************************************************
   * construct typed identifier from identifier and type.
   * @param identifier the identifier
   * @param type its type
   ***************************************************************************/
  public TypedIdentifier(ValueDeclIdentifier identifier, Type type)
  {
    this.identifier = identifier;
    this.type = type;
  }
  
  /****************************************************************************
   * returns identifier name
   * @return the identifier name
   ***************************************************************************/
  public ValueDeclIdentifier getIdentifier()
  {
    return identifier;
  }
  
  /****************************************************************************
   * returns identifier type
   * @return the identifier type
   ***************************************************************************/
  public Type getType()
  {
    return type;
  }
  
  /***************************************************************************
   * create instantiated copy of identifier
   * @return the substitution
   **************************************************************************/
  public TypedIdentifier instantiate()
  {
    ValueDeclIdentifier ident = getIdentifier().instantiate();
    Type type = getType().instantiateType();
    return Construct.typedIdentifier(ident, type);
  }
  
  /****************************************************************************
   * set substitution expression for instantiation
   * @param substitution the substitution expression
   ***************************************************************************/
  public void setSubstitution(TypedIdentifier substitution)
  {
    getIdentifier().setSubstitution(substitution.getIdentifier());
  }
  
  /***************************************************************************
   * reset substitution field in identifier
   **************************************************************************/
  public void resetSubstitution()
  {
    getIdentifier().resetSubstitution();
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    identifier.print(out);
    out.print(":");
    type.print(out);
  }
  
  /***************************************************************************
   * Prints text representation. Overrides ASTBase.print
   * such that parentheses are never printed around typed identifiers.
   * @param out the stream on which the text is written
   **************************************************************************/
  public void print(PrintWriter out)
  {
    printCore(out);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST identifier0 = identifier.accept(visitor);
    AST type0 = type.accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.typedIdentifier(
      (ValueDeclIdentifier)identifier0, (Type)type0);
  }
}
